# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Definitions that are used by both users and user groups services."""

from invenio_records_resources.services import Link as LinkBase


class Link(LinkBase):
    """Shortcut for writing links with IDs."""

    @staticmethod
    def vars(obj, vars):
        """Variables for the URI template."""
        vars.update({"id": obj.id})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module providing management APIs for users and roles/groups."""

from flask import Blueprint

blueprint = Blueprint(
    "invenio_users_resources",
    __name__,
    template_folder="templates",
    static_folder="static",
)


@blueprint.record_once
def init(state):
    """Init app."""
    app = state.app
    # Register services - cannot be done in extension because
    # Invenio-Records-Resources might not have been initialized.
    rr_ext = app.extensions["invenio-records-resources"]
    ext = app.extensions["invenio-users-resources"]
    idx_ext = app.extensions["invenio-indexer"]

    # services
    rr_ext.registry.register(ext.users_service)
    rr_ext.registry.register(ext.groups_service)
    rr_ext.registry.register(ext.domains_service)

    idx_ext.registry.register(ext.users_service.indexer, indexer_id="users")
    idx_ext.registry.register(ext.groups_service.indexer, indexer_id="groups")
    idx_ext.registry.register(ext.domains_service.indexer, indexer_id="domains")


def create_users_resources_bp(app):
    """Create the users resource blueprint."""
    ext = app.extensions["invenio-users-resources"]
    return ext.users_resource.as_blueprint()


def create_groups_resources_bp(app):
    """Create the user groups resource blueprint."""
    ext = app.extensions["invenio-users-resources"]
    return ext.groups_resource.as_blueprint()


def create_domains_resources_bp(app):
    """Create the domains resource blueprint."""
    ext = app.extensions["invenio-users-resources"]
    return ext.domains_resource.as_blueprint()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Filter parameter interpreter API."""

from functools import partial

from invenio_records_resources.services.records.params import ParamInterpreter
from invenio_search.engine import dsl


class ModerationFilterParam(ParamInterpreter):
    """Evaluate type filter."""

    def __init__(self, param_name, field_name, config):
        """."""
        self.param_name = param_name
        self.field_name = field_name
        super().__init__(config)

    @classmethod
    def factory(cls, param=None, field=None):
        """Create a new filter parameter."""
        return partial(cls, param, field)

    def apply(self, identity, search, params):
        """Applies a filter to get only records for a specific type."""
        # Pop because we don't want it to show up in links.
        value = params.pop(self.param_name, None)
        # TODO this can be broken down to separate param components
        if value is not None:
            if self.param_name == "is_blocked" or self.param_name == "is_verified":
                if value is True:
                    search = search.filter(
                        dsl.query.Bool(
                            **{"must": {"exists": {"field": self.field_name}}}
                        )
                    )
                else:
                    search = search.filter(
                        dsl.query.Bool(
                            **{"must_not": {"exists": {"field": self.field_name}}}
                        )
                    )
            elif self.param_name == "is_active":
                if value is True:
                    search = search.filter("term", **{self.field_name: True})
                else:
                    # search only active but not blocked
                    search = search.filter(
                        dsl.query.Q("term", **{self.field_name: False})
                        & dsl.query.Bool(
                            **{"must_not": {"exists": {"field": "blocked_at"}}}
                        )
                    )

        return search



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Domains resource tests."""

import pytest


def test_domains_access(app, client, domains, user_moderator, user_pub):
    res = client.get(f"/domains")
    assert res.status_code == 403

    user_pub.login(client)
    res = client.get(f"/domains")
    assert res.status_code == 403


def test_domains_search(app, client, domains, user_moderator, user_pub):
    user_moderator.login(client)
    res = client.get(f"/domains")
    assert res.status_code == 200
    data = res.json
    assert len(data["hits"]["hits"]) == 5

    # Props
    props = [
        "id",
        "domain",
        "created",
        "updated",
        "domain",
        "tld",
        "status",
        "status_name",
        "category",
        "category_name",
        "flagged",
        "flagged_source",
        "org",
        "num_users",
        "num_active",
        "num_inactive",
        "num_confirmed",
        "num_verified",
        "num_blocked",
    ]
    cern = data["hits"]["hits"][0]
    assert cern["domain"] == "cern.ch"
    for p in props:
        assert p in cern

    # Check aggregations and that they have content
    aggs = ["status", "flagged", "category", "organisation", "tld"]
    for a in aggs:
        assert a in data["aggregations"]
        assert (
            len(data["aggregations"][a]["buckets"]) > 0
        ), f"'{a}' is missing bucket values"
    assert len(data["hits"]["hits"]) == 5


def test_domains_read(app, client, domains, user_moderator):
    res = client.get(f"/domains/cern.ch")
    assert res.status_code == 403

    user_moderator.login(client)
    res = client.get(f"/domains/cern.ch")
    assert res.json["links"]["self"].endswith("/domains/cern.ch")
    assert res.status_code == 200
    d = res.json
    assert d["domain"] == "cern.ch"
    assert d["tld"] == "ch"
    assert d["status"] == 3
    assert d["status_name"] == "verified"
    assert d["flagged"] == False
    assert d["flagged_source"] == ""
    assert d["category"] == 1
    assert d["category_name"] == "organization"
    assert d["org"] == [
        {
            "id": 1,
            "pid": "https://ror.org/01ggx4157",
            "name": "CERN",
            "props": {"country": "ch"},
            "is_parent": False,
        }
    ]
    stats = ["users", "active", "inactive", "confirmed", "verified", "blocked"]
    for s in stats:
        assert f"num_{s}" in d, f"num_{s} is missing from payload"


def test_domains_delete(app, client, domains, user_moderator):
    res = client.delete(f"/domains/cern.ch")
    assert res.status_code == 403

    user_moderator.login(client)
    res = client.delete(f"/domains/cern.ch")
    assert res.status_code == 204
    res = client.get(f"/domains/cern.ch")
    assert res.status_code == 404


def test_domains_create(app, client, domains, user_moderator):
    res = client.post(
        f"/domains",
        json={
            "domain": "zenodo.org",
        },
        headers={"content-type": "application/vnd.inveniordm.v1+json"},
    )
    assert res.status_code == 403

    user_moderator.login(client)
    # Make an update
    res = client.post(
        f"/domains",
        json={
            "domain": "zenodo.org",
        },
        headers={"content-type": "application/vnd.inveniordm.v1+json"},
    )
    assert res.status_code == 201
    # Re-read to check that it was updated
    data = client.get(f"/domains/zenodo.org").json
    assert data["domain"] == "zenodo.org"
    assert data["tld"] == "org"
    assert data["status_name"] == "new"
    assert data["category_name"] == None
    assert data["flagged"] == False
    assert data["flagged_source"] == ""
    assert data["org"] is None


@pytest.mark.parametrize(
    "status_code,json",
    [
        (400, {"status": "new"}),  # missing domain
        (400, {"domain": "test.com", "status_name": "invalid"}),  # invalid status
        (400, {"domain": "spammer.com"}),  # duplicate domain
        (400, {"domain": "test.com", "category_name": "invalid"}),  # invalid category
    ],
)
def test_domains_create_failure(
    app, client, domains, user_moderator, status_code, json
):
    user_moderator.login(client)
    # Make an update
    res = client.post(
        f"/domains",
        json=json,
        headers={"content-type": "application/vnd.inveniordm.v1+json"},
    )
    assert res.status_code == status_code


def test_domains_update(app, client, domains, user_moderator):
    user_moderator.login(client)
    data = client.get(f"/domains/moderated.org").json
    assert data["domain"] == "moderated.org"
    assert data["status_name"] == "moderated"
    assert data["category_name"] == "mail-provider"
    assert data["flagged"] == True
    assert data["flagged_source"] == "disposable"
    assert data["org"] is None
    # Make an update
    res = client.put(
        f"/domains/moderated.org",
        json={
            "domain": "moderated.org",
            "status_name": "verified",
            "category_name": "spammer",
            "flagged": False,
            "flagged_source": "test",
            "org": None,
        },
        headers={"content-type": "application/vnd.inveniordm.v1+json"},
    )
    assert res.status_code == 200
    # Re-read to check that it was updated
    data = client.get(f"/domains/moderated.org").json
    assert data["domain"] == "moderated.org"
    assert data["tld"] == "org"
    assert data["status_name"] == "verified"
    assert data["category_name"] == "spammer"
    assert data["flagged"] == False
    assert data["flagged_source"] == "test"
    assert data["org"] is None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Base model classes for user and group management in Invenio."""

from abc import ABC, abstractmethod

from flask import current_app
from invenio_accounts.proxies import current_datastore
from invenio_accounts.utils import DomainStatus
from invenio_db import db


class AggregateMetadata(ABC):
    """Model class that does not corresponds to a database table.

    Since we already have all information about the required entities
    stored in the database and just want to provide a central API to
    manage these entities as one, we do not want to store any of the
    information in the database again.
    Instead, we provide a mock model that provides the model interface
    that's expected by API classes as far as necessary.
    """

    _properties = []
    """Properties of this object that can be accessed."""

    _set_properties = []
    """Properties of this object that can be set."""

    _data = None

    def __init__(self, model_obj=None, **kwargs):
        """Constructor."""
        super().__setattr__("_data", {})
        if model_obj is not None:
            # E.g. when data is loaded from database
            self.from_model(model_obj)
            super().__setattr__("_model_obj", model_obj)
        else:
            # E.g. when data is loaded from the search index
            self.from_kwargs(kwargs)
            super().__setattr__("_model_obj", None)

    @property
    @abstractmethod
    def model_obj(self):
        """The actual model object behind this mock model."""
        return None

    def from_kwargs(self, kwargs):
        """Extract information from kwargs."""
        for p in self._properties:
            self._data[p] = kwargs.get(p, None)

    def from_model(self, model_obj):
        """Extract information from a user/role object."""
        # Edit self._properties if you need to add more properties
        for p in self._properties:
            self._data[p] = getattr(model_obj, p, None)

    def __getattr__(self, name):
        """Get an attribute from the model."""
        if name in self._properties:
            return self._data[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        """Set an attribute from the model."""
        if name not in self._set_properties:
            raise AttributeError(name)
        super().__setattr__(name, value)
        setattr(self.model_obj, name, value)

    # Methods required to make it a record.
    @property
    def is_deleted(self):
        """Method needed for the record API."""
        return False

    @property
    def json(self):
        """Method needed for the record API."""
        return {p: getattr(self, p, None) for p in self._properties}

    @property
    def data(self):
        """Method needed for the record API."""
        return {p: getattr(self, p, None) for p in self._properties}


class UserAggregateModel(AggregateMetadata):
    """User aggregate data model."""

    # If you add properties here you likely also want to add a ModelField on
    # the UserAggregate API class.
    _properties = [
        "id",
        "version_id",
        "email",
        "domain",
        "username",
        "active",
        "preferences",
        "profile",
        "confirmed_at",
        "blocked_at",
        "verified_at",
        "created",
        "updated",
        "current_login_at",
    ]
    """Properties of this object that can be accessed."""

    _set_properties = [
        "active",
        "blocked_at",
        "confirmed_at",
        "verified_at",
    ]
    """Properties of this object that can be set."""

    def from_model(self, user):
        """Extract information from a user object."""
        super().from_model(user)
        self._data["profile"] = dict(user.user_profile or {})

        # Set defaults
        self._data["preferences"] = dict(self._data["preferences"] or {})
        self._data["preferences"].setdefault("visibility", "restricted")
        self._data["preferences"].setdefault("email_visibility", "restricted")
        default_locale = current_app.config.get("BABEL_DEFAULT_LOCALE", "en")
        self._data["preferences"].setdefault("locale", default_locale)
        self._data["preferences"].setdefault("timezone", "Europe/Zurich")
        self._data["preferences"].setdefault(
            "notifications",
            {
                "enabled": True,
            },
        )

    @property
    def model_obj(self):
        """The actual model object behind this user aggregate."""
        if self._model_obj is None:
            id_ = self._data.get("id")
            with db.session.no_autoflush:
                self._model_obj = current_datastore.get_user_by_id(id_)
        return self._model_obj


class GroupAggregateModel(AggregateMetadata):
    """Mock model for glueing together various parts of user group info."""

    _properties = [
        "id",
        "version_id",
        "name",
        "description",
        "is_managed",
        "created",
        "updated",
    ]
    """Properties of this object that can be accessed."""

    _set_properties = []
    """Properties of this object that can be set."""

    @property
    def model_obj(self):
        """The actual model object behind this mock model."""
        if self._model_obj is None:
            name = self.data.get("id")
            with db.session.no_autoflush:
                self._model_obj = current_datastore.find_role(name)
        return self._model_obj


class DomainAggregateModel(AggregateMetadata):
    """Mock model for glueing together various parts of user group info."""

    _properties = [
        "category",
        "created",
        "domain",
        "flagged_source",
        "flagged",
        "id",
        "num_active",
        "num_blocked",
        "num_confirmed",
        "num_inactive",
        "num_users",
        "num_verified",
        "org_id",
        "status",
        "tld",
        "updated",
        "version_id",
    ]
    """Properties of this object that can be accessed."""

    _set_properties = [
        "category",
        "domain",
        "flagged_source",
        "flagged",
        "org_id",
        "status",
        "tld",
    ]
    """Properties of this object that can be set."""

    def from_model(self, domain):
        """Extract information from a user object."""
        super().from_model(domain)
        # Hardcoding version id to 1 since domain model doesn't have
        # a version id because we update the table often outside
        # of sqlalchemy ORM.
        self._data["version_id"] = 1
        # Convert enum
        status = self._data.get("status", None)
        if status and isinstance(status, DomainStatus):
            self._data["status"] = status.value

    @property
    def model_obj(self):
        """The actual model object behind this mock model."""
        if self._model_obj is None:
            domain = self.data.get("domain")
            with db.session.no_autoflush:
                self._model_obj = current_datastore.find_domain(domain)
        return self._model_obj
